'
'  pb_tcpip.bas
'
'  tcp/ip routines for pb/dll and pb/cc
'
'  Written by Don Dickinson - ddickinson@usinternet.com
'  July, 1998
'
'  Hereby Public Domain. Deemed reliable, but use at your own risk, etc.
'  Do with it as you will, provided you hold the author harmless
'  from all effects and side-effects of using this code.
'
#if not %def(%PB_TCPIP_BAS)
%PB_TCPIP_BAS = 1

#if not %def(%WINAPI)
#include "win32api.inc"
#endif
#include "wsock32.inc"
#include "pb_doeve.bas"

Global g_wp_Window as Long
Global g_wp_Init as Long

declare function callProgress_TEMPLATE ( byval iCount as long, _
      byval iTotal as long ) as long

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  Debug Flags
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
%DEBUG_TCPIP_CONSOLE          = %false
%DEBUG_TCPIP_GUI              = %false

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  Constants
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
%PB_SOCKET_TIMEOUT = 60000        '60 second timeout

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipInit
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipInit alias "ipInit" export as long

   Dim rData as WSAData

   if g_wp_Init then
      function = %true
   elseif WSAStartup(&H0101, rData) = -1 then
      function = %false
   else
      g_wp_Init = %true
      function = %true
   end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipShutdown
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub ipShutdown alias "ipShutdown" export

   WSACleanup

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dottedToIP
'
'  Takes a dotted ip address and returns the integer form of the address.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function dottedToIP alias "dottedToIP" _
      ( ByVal sHostName as String ) export as Long

   Dim i as Long
   Dim iReturn as Long
   Dim onePart(1 to 4) as Long
   Dim pBuf as Byte Ptr

   if tally(sHostName, ".") <> 3 then
      function = 0
   else
      onePart(1) = int(val(parse$(sHostName, ".", 1)))
      onePart(2) = int(val(parse$(sHostName, ".", 2)))
      onePart(3) = int(val(parse$(sHostName, ".", 3)))
      onePart(4) = int(val(parse$(sHostName, ".", 4)))

      pBuf = varptr(iReturn)
      for i =1 to 4
         if (onePart(i) < 0) or (onePart(i) > 255) then
            function = 0
            exit function
         end if

         @pBuf = asc(chr$(onePart(i)))
         incr pBuf
      next i

   end if
   function = iReturn

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipIsDotted
'
'  Returns %true if the host name is represented in dotted-ip fasion.
'  Return %false if not. e.g.
'
'  ipIsDotted("www.powerbasic.com") returns %false
'  ipIsDotted("192.145.12.1") returns %true
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipIsDotted alias "ipIsDotted" _
      ( ByVal sHostName as String ) export as Long

   Dim i as Long
   Dim iZeroCount as Long
   Dim onePart(1 to 4) as Long

   if tally(sHostName, ".") <> 3 then
      function = %false
   else
      onePart(1) = int(val(parse$(sHostName, ".", 1)))
      onePart(2) = int(val(parse$(sHostName, ".", 2)))
      onePart(3) = int(val(parse$(sHostName, ".", 3)))
      onePart(4) = int(val(parse$(sHostName, ".", 4)))

      iZeroCount = 0
      for i =1 to 4
         if onePart(i) = 0 then
            incr iZeroCount
         elseif (onePart(i) < 0) or (onePart(i) > 255) then
            function = %false
            exit function
         end if
      next i

      if iZeroCount > 3 then
         function = %false
      else
         function = %true
      end if
   end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipGetAddress
'
'  Returns the ip address of the given host in big endian form
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipGetAddress alias "ipGetAddress" _
      ( ByVal sHostName as String ) export as Long

   Dim ptrHost As hostentStru Ptr

   if ipIsDotted(sHostName) then
      function = dottedToIP(sHostName)
   else
      ptrHost = gethostbyname(sHostName + $nul)
      If ptrHost Then
         Function = @ptrHost.@@h_list
      Else
         Function = 0
      End If
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipGetHostAddress
'
'  Fills a sockaddr_in structure with the port and address information
'  given the host amd port
'
'  e.g. ipGetHostAddress "www.winrep.com", 80, rSock
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipGetHostAddress Alias "ipGetHostAddress" _
      ( ByVal sHost as String, ByVal iPort as Integer, _
        rSockAddr As sockaddr_in) Export as Long

   Dim iReturn As Integer
   Dim ptrHost As hostentStru Ptr
   Dim ptrServeEnt As servent Ptr

   rSockAddr.sin_family = %AF_INET
   rSockAddr.sin_addr.s_addr = ipGetAddress(sHost)
   if rSockAddr.sin_addr.s_addr = 0 then
      $If %DEBUG_TCPIP_CONSOLE
      Print "gethostbyname failed"
      $ElseIf %DEBUG_TCPIP_GUI
      MsgBox "gethostbyname failed"
      $EndIf
      Function = %False
   Else
      rSockAddr.sin_port = htons(iPort)
      Function = %True
   End If

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceive
'
'  You can't just call the rrecv function and assume that it sent
'  you all of the information you wanted, you need to keep calling
'  rrecv until it returns nothing - this "packet fragmenting" is a
'  result of the microsoft implementation of tcp/ip.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceive alias "ipSafeReceive" _
      ( ByVal iSocket As Long, byval iTimeout as long ) export As String

   Dim iCount As Integer
   Dim iResult As Long
   Dim zInBuffer As String * 4096
   Dim sAccum As String

   $If %DEBUG_TCPIP_CONSOLE
   Print "===== ipSafeRecv ====="
   $ElseIf %DEBUG_TCPIP_GUI
   MsgBox "Starting ipSafeRecv"
   $EndIf

   sAccum = ""
   iCount = 0

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if

   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iResult = rrecv(iSocket, zInBuffer, SizeOf(zInBuffer), 0)
   Do Until iResult < 1
      Incr iCount
      $If %DEBUG_TCPIP_CONSOLE
      Stdout "RECV: " + format$(iCount) + _
             " res:" + format$(iResult) + _
             " len=" + format$(Len(sAccum))

      Print Left$(zInBuffer, iResult)
      $ElseIf %DEBUG_TCPIP_GUI
      MsgBox   "RECV" + $crlf + _
               "iCount =" + Str$(iCount) + Chr$(13) + _
               "iResult =" + Str$(iResult) + Chr$(13) + _
               "len(sAccum) =" + Str$(Len(sAccum)) + Chr$(13) + _
               "zInBuffer: " + Chr$(13) + Trim$(zInBuffer)
      $EndIf
      sAccum = sAccum + Left$(zInBuffer, iResult)
      if g_wp_Window then
         SetWindowText g_wp_Window, format$(len(sAccum)) + " bytes received."
         apiSleep 1
      end if
      iResult = rrecv(iSocket, zInBuffer, SizeOf(zInBuffer), 0)
   Loop
   if g_wp_Window then
      SetWindowText g_wp_Window, format$(len(sAccum)) + " bytes received."
      apiSleep 1
   end if

   'write_log "iResult=" + format$(iResult) + " " + hex$(iResult)

   $If %DEBUG_TCPIP_CONSOLE
   Print "iResult =" + Str$(iResult)
   Print "WSAGetLastError =" + Str$(WSAGetLastError)
   Print "===== END ipSafeReceive ====="
   $ElseIf %DEBUG_TCPIP_GUI
   MsgBox "iResult =" + Str$(iResult)
   $EndIf

   Function = sAccum

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceiveUntil
'
'  Received data until the sUntil string is found in the input
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceiveUntil alias "ipSafeReceiveUntil" _
      ( ByVal iSocket As Long, ByVal sUntil as String, _
        byval iTimeout as long ) export As String

   Dim iCount As Integer
   Dim iResult As Long
   Dim zInBuffer As String * 1
   Dim sAccum As String

   sAccum = ""
   iCount = 0

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if

   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iResult = rrecv(iSocket, zInBuffer, SizeOf(zInBuffer), 0)
   Do Until iResult < 1
      Incr iCount
      sAccum = sAccum + zInBuffer
      if instr(sAccum, sUntil) then exit do
      iResult = rrecv(iSocket, zInBuffer, SizeOf(zInBuffer), 0)
   Loop

   Function = sAccum

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceiveBuffer
'
'  Receives the number of bytes requested
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceiveBuffer alias "ipSafeReceiveBuffer" _
      ( ByVal iSocket As Long, ByVal iLen as Long, _
        byval iTimeout as long ) export As String

   Dim iResult As Long
   Dim iReturnLen as Long
   Dim iLeft as Long
   Dim zInBuffer As String
   Dim sAccum As String

   sAccum = ""

   zInBuffer = space$(iLen)
   iReturnLen = iLen

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if

   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iLeft = iLen
   iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Do Until iResult < 1
      sAccum = sAccum + left$(zInBuffer, iResult)
      if len(sAccum) >= iLen then exit do
      iLeft = iLeft - iResult
      apiSleep 1
      iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Loop

   Function = sAccum

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceiveBuffer2file
'
'  Receives the number of bytes requested to a file.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceiveBuffer2file alias "ipSafeReceiveBuffer2file" _
      ( ByVal iSocket As Long, ByVal iLen as Long, _
        byval iTimeout as long, byval hFile as long) export As long

   Dim iResult As Long
   Dim iReturnLen as Long
   Dim iLeft as Long
   dim iRecv as long
   Dim zInBuffer As String
   dim sBuf as string

   zInBuffer = space$(iLen)
   iReturnLen = iLen

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if

   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iLeft = iLen
   iRecv = 0
   iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Do Until iResult < 1
      sBuf = left$(zInBuffer, iResult)
      put #hFile,, sBuf
      iRecv = iRecv + iResult
      if iRecv >= iLen then exit do
      apiSleep 1
      iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Loop

   Function = iRecv

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceiveBuffer2
'
'  Receives the number of bytes requested. Sends status message
'  to the hWnd control via setWindowText
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceiveBuffer2 alias "ipSafeReceiveBuffer2" _
      ( ByVal iSocket As Long, ByVal iLen as Long, _
        byval iTimeout as long, byval hWnd as long ) export As String

   dim iResult As Long
   dim iReturnLen as Long
   dim iLeft as Long
   dim zInBuffer As String
   dim sAccum As String

   sAccum = ""
   zInBuffer = space$(iLen)
   iReturnLen = iLen

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if
   '!!!
   'iTimeout = 0
   if iTimeout > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iLeft = iLen
   iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Do Until iResult < 1
      sAccum = sAccum + left$(zInBuffer, iResult)
      if len(sAccum) >= iLen then exit do
      iLeft = iLeft - iResult
      if hWnd <> 0 then
         setWindowText hWnd, "received " + format$(len(sAccum)) + _
               " of " + format$(iLen)
         apiSleep 1

      end if
      iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Loop

   Function = sAccum

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceiveBuffer2
'
'  Receives the number of bytes requested. Sends status message
'  to the hWnd control via setWindowText
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceiveBuffer22file alias "ipSafeReceiveBuffer22file" _
      ( ByVal iSocket As Long, ByVal iLen as Long, _
        byval iTimeout as long, byval hWnd as long, _
        byval hFile as long ) export As long

   dim iResult As Long
   dim iReturnLen as Long
   dim iLeft as Long
   dim iRecv as long
   dim zInBuffer As String
   dim sBuf as string

   zInBuffer = space$(iLen)
   iReturnLen = iLen

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if
   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iLeft = iLen
   iRecv = 0
   iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Do Until iResult < 1
      sBuf = left$(zInBuffer, iResult)
      put #hFile,, sBuf
      iRecv = iRecv + iResult
      if hWnd <> 0 then
         setWindowText hWnd, "received " + format$(iRecv) + _
               " of " + format$(iLen)
         apiSleep 1

      end if
      if iRecv >= iLen then exit do
      iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Loop

   Function = iRecv

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceiveBuffer3
'
'  Receives the number of bytes requested. Sends status message
'  to the code ptr = ptrProgress
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceiveBuffer3 alias "ipSafeReceiveBuffer3" _
      ( ByVal iSocket As Long, ByVal iLen as Long, _
        byval iTimeout as long, byval ptrProgress as dword ) export As String

   dim iResult As Long
   dim iReturnLen as Long
   dim iLeft as Long
   dim zInBuffer As String
   dim sAccum As String

   sAccum = ""
   zInBuffer = space$(iLen)
   iReturnLen = iLen

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if
   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iLeft = iLen
   iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Do Until iResult < 1
      sAccum = sAccum + left$(zInBuffer, iResult)
      if len(sAccum) >= iLen then exit do
      iLeft = iLeft - iResult
      if ptrProgress <> 0 then
         call dword ptrProgress _
               using callProgress_TEMPLATE ( len(sAccum), iLen )
         apiSleep 1
      end if
      iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Loop

   Function = sAccum

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeReceiveBuffer32file
'
'  Receives the number of bytes requested. Sends status message
'  to the code ptr = ptrProgress
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipSafeReceiveBuffer32file alias "ipSafeReceiveBuffer32file" _
      ( ByVal iSocket As Long, ByVal iLen as Long, _
        byval iTimeout as long, byval ptrProgress as dword, _
        byval hFile as long ) export As long

   dim iResult As Long
   dim iReturnLen as Long
   dim iLeft as Long
   dim iRecv as long
   dim zInBuffer As String
   dim sBuf as string

   zInBuffer = space$(iLen)
   iReturnLen = iLen

   '- How long will we wait (limits blocking)
   if iTimeout = 0 then
      iTimeout = %PB_SOCKET_TIMEOUT
   end if
   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iLeft = iLen
   iRecv = 0
   iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Do Until iResult < 1
      sBuf = left$(zInBuffer, iResult)
      put #hFile,, sBuf
      iRecv = iRecv + iResult
      if ptrProgress <> 0 then
         call dword ptrProgress _
               using callProgress_TEMPLATE ( iRecv, iLen )
         apiSleep 1

      end if
      if iRecv >= iLen then exit do
      iResult = rrecv(iSocket, ByVal strptr(zInBuffer), iLeft, 0)
   Loop

   function = iRecv

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipReceiveLine
'
'  This is just like ipSafeReceive except that it quits receiving
'  after it gets a chr$(13,10). This is used with pop3 communications
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipReceiveLine alias "ipReceiveLine" _
      ( ByVal iSocket As Long ) export as String

   Dim iCount As Integer
   Dim iResult As Long
   Dim iTimeout As Long
   Dim zInBuffer As String * 1
   Dim sAccum As String

   $If %DEBUG_TCPIP_CONSOLE
   Print "===== ipReceiveLine ====="
   $ElseIf %DEBUG_TCPIP_GUI
   MsgBox "ipReceiveLine starting"
   $EndIf

   sAccum = ""
   iCount = 0

   '- How long will we wait (limits blocking)
   iTimeout = %PB_SOCKET_TIMEOUT
   '!!!
   'iTimeout = 0
   if iTimeOut > 0 then
      setsockopt iSocket, %SOL_SOCKET, %SO_RCVTIMEO, iTimeout, SizeOf(iTimeout)
   end if
   iResult = rrecv(iSocket, ByVal VarPtr(zInBuffer), SizeOf(zInBuffer), 0)
   Do Until iResult < 1
      Incr iCount
      sAccum = sAccum + zInBuffer
      If right$(sAccum, 2) = $crlf then exit do
      iResult = rrecv(iSocket, zInBuffer, SizeOf(zInBuffer), 0)
   Loop

   Function = sAccum

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipReceiveToFile
'
'  Retrieves data exactly as ipSafeSend does, except that it
'  echos the input to a file. If the file can't be opened or
'  created (if it already exists, for example) it returns %false.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function ipReceiveToFile Alias "ipReceiveToFile" _
      ( ByVal iSocket As Long, sFileName As String ) Export As String

   Dim iFF As Long
   Dim iCount As Integer
   Dim iResult As Long
   Dim zInBuffer As String * 4097
   Dim sAccum As String

   On Error Resume Next
   If Dir$(sFileName) <> "" Then
      iFF = 0
   else
      iFF = FreeFile
      Open sFileName For Binary As #iFF
      If Err Then
         iFF = 0
      End If
   end if

   sAccum = ""
   iCount = 0

   iResult = rrecv(iSocket, ByVal VarPtr(zInBuffer), SizeOf(zInBuffer) - 1, 0)
   Do Until iResult < 1
      Incr iCount
      $If %DEBUG_TCPIP_CONSOLE
      Print iCount, iResult, Len(sAccum)
      $ElseIf %DEBUG_TCPIP_GUI
      MsgBox   "iCount =" + Str$(iCount) + Chr$(13) + _
               "iResult =" + Str$(iResult) + Chr$(13) + _
               "len(sAccum) =" + Str$(Len(sAccum)) + Chr$(13) + _

      $EndIf
      sAccum = sAccum + Left$(zInBuffer, iResult)
      iResult = rrecv(iSocket, ByVal VarPtr(zInBuffer), SizeOf(zInBuffer) - 1, 0)
   Loop

   If iFF Then
      Put #iFF,, sAccum
      Close iFF
   end if
   function = sAccum

End Function

#if 0

ORIGINAL VERSION
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeSend
'
'  Sends data over an open socket
'  returns the number of bytes sent.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipSafeSend alias "ipSafeSend" _
      ( ByVal iSocket As Long, sData As String ) Export As Long

   function = ssend(iSocket, ByVal StrPtr(sData), Len(sData), 0)

end function
#endif

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipSafeSend
'
'  Sends data over an open socket
'  returns the number of bytes sent.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipSafeSend alias "ipSafeSend" _
      ( ByVal iSocket As Long, sData As String ) Export As Long

   dim iSent as long
   dim iTotal as long
   dim sToSend as string

   iTotal = 0
   iSent = ssend(iSocket, byval strptr(sData), Len(sData), 0)

   do
      iTotal = iTotal + iSent
      if iSent < 0 then
         iTotal = -1
         exit do
      elseif iTotal >= len(sData) then
         exit do
      end if

      sToSend = mid$(sData, iTotal + 1)
      iSent = ssend(iSocket, byval strptr(sToSend), len(sToSend), 0)

   loop

'   if iTotal < len(sData) then
'      msgbox "send error. Tried to send " + format$(len(sData)) + _
'            " bytes, but only really sent " + format$(iTotal) + "."
'   end if

   function = iTotal

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipIsConnected
'
'  Checks to see if we need to reestablish a connection
'  with the server - if the socket doesn't work.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipIsConnected Alias "ipIsConnected" _
      (ByVal iSocket As Long) export as Long

   Dim rPeer As sockaddr_in
   Dim iNameLen As Long

   iNamelen = Len(sockaddr)
   If getpeername(iSocket, rPeer, iNameLen) = 0 Then
      Function = %True
   Else
      Function = %False
   End If

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipToDotted
'
'  This is a silly utility that gets an IP address
'  for a given host and prints it to the screen.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipToDotted Alias "ipToDotted" _
      ( ByVal sHost as String ) export as String

   Dim iHost As Long
   Dim sLong As String

   iHost = ipGetAddress(sHost)
   sLong = Mkl$(iHost)
   function = Format$(Asc(Left$(sLong, 1))) + "." + _
              Format$(Asc(Mid$(sLong,2,1))) + "." + _
              Format$(Asc(Mid$(slong,3,1))) + "." + _
              Format$(Asc(Right$(slong,1)))

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  ipConnect
'  Connects to a server on specified socket.
'  Returns %INVALID_SOCKET (-1) on failure or
'  the new socket's handle on success.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function ipConnect Alias "ipConnect" _
      ( ByVal sServer as String, ByVal iPort as Long ) export as long

   Dim iSocket As Long
   Dim iReturn As Long
   Dim rSockAddr As sockaddr_in

   if g_wp_Init = %false then
      ipInit
      g_wp_Init = %true
   end if

   if ipGetHostAddress(sServer, iPort, rSockAddr) = %false then
      function = %INVALID_SOCKET
      exit function
   end if

   '- create the socket
   iSocket = socket(%AF_INET, %SOCK_STREAM, 0)
   If iSocket = 0 Then
      Function = %INVALID_SOCKET
      exit function
   end if

   '- Connect to the server
   if cconnect(iSocket, rSockAddr, Len(rSockAddr)) = %SOCKET_ERROR then
      closesocket iSocket
      function = %INVALID_SOCKET
   else
      function = iSocket
   end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  httpSendAndReceive
'
'  This function takes some data and sends it off to a server. Then it waits
'  for the server's response and puts the data the server sent back in
'  the sRecvBuffer string.
'  It returns %True on success and %False on failure.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function httpSendAndReceive Alias "httpSendAndReceive" _
      (  ByVal sServer as String, ByVal sSendBuffer As String, _
         sRecvBuffer As String ) Export as Long

   Dim iSocket As Long

   '- Make the connection
   iSocket = ipConnect(sServer, 80)
   if iSocket = %INVALID_SOCKET then
      function = %false
      exit function
   end if

   '- Send the data
   If ipSafeSend(iSocket, sSendBuffer) < Len(sSendBuffer) Then
      sRecvBuffer = ""
      shutdown iSocket, %SD_SEND
      closesocket iSocket
      Function = %False
      exit function
   end if

   '- Retrieve the data
   sRecvBuffer = ipSafeReceive(iSocket, %PB_SOCKET_TIMEOUT)

   '- Clean up
   shutdown iSocket, %SD_SEND
   closesocket iSocket

   Function = %True

End Function

#endif
